/*-------------------------------------------------------------------------
 *
 * deparse.c
 *		  query deparser for PostgreSQL
 *
 * Copyright (c) 2011, PostgreSQL Global Development Group
 *
 * IDENTIFICATION
 *		  pgsql_fdw/deparse.c
 *
 *-------------------------------------------------------------------------
 */
#include "postgres.h"

#include "access/transam.h"
#include "foreign/foreign.h"
#include "lib/stringinfo.h"
#include "nodes/nodeFuncs.h"
#include "nodes/nodes.h"
#include "nodes/makefuncs.h"
#include "optimizer/clauses.h"
#include "optimizer/var.h"
#include "parser/parsetree.h"
#include "utils/builtins.h"
#include "utils/lsyscache.h"

#include "pgsql_fdw.h"

/*
 * Context for walk-through the expression tree.
 */
typedef struct foreign_executable_cxt
{
	PlannerInfo	   *root;
	RelOptInfo	   *foreignrel;
} foreign_executable_cxt;

static bool is_foreign_expr(PlannerInfo *root, RelOptInfo *baserel, Expr *expr);
static bool foreign_expr_walker(Node *node, foreign_executable_cxt *context);
static bool is_proc_remotely_executable(Oid procid);
static void append_where_clause(StringInfo buf, List *context, Path *path,
								bool *first);

/*
 * Store qualified relation name into result.
 */
static void
getQualifiedRelationName(Oid relid, StringInfo result)
{
	ForeignTable   *ft;				/* foreign table to be scanned */
	const char *nspname = NULL;		/* plain namespace name */
	const char *relname = NULL;		/* plain relation name */
	const char *q_nspname;			/* quoted namespace name */
	const char *q_relname;			/* quoted relation name */

	ft = GetForeignTable(relid);
	if (ft->options != NIL)
	{
		ListCell	   *lc;

		foreach (lc, ft->options)
		{
			DefElem	   *opt = lfirst(lc);
			if (strcmp(opt->defname, "nspname") == 0)
				nspname = strVal(opt->arg);
			else if (strcmp(opt->defname, "relname") == 0)
				relname = strVal(opt->arg);
		}
	}
	if (nspname == NULL)
		nspname = get_namespace_name(get_rel_namespace(relid));
	if (relname == NULL)
		relname = get_rel_name(relid);
	q_nspname = quote_identifier(nspname);
	q_relname = quote_identifier(relname);
	appendStringInfo(result, "%s.%s", q_nspname, q_relname);
}

/*
 * Deparse query representation into SQL statement which suits for remote
 * PostgreSQL server.  Also some of quals in WHERE clause will be pushed down
 * If they are safe to be evaluated on the remote side.
 */
char *
deparseSql(Oid relid, PlannerInfo *root, RelOptInfo *baserel)
{
	StringInfoData	foreign_relname;
	StringInfoData	sql;			/* builder for SQL statement */
	bool		first;
	AttrNumber	attr;
	List	   *attr_used = NIL;	/* List of AttNumber used in the query */
	List	   *foreign_expr = NIL;	/* list of Expr* evaluated on remote */
	int			i;
	List	   *rtable = NIL;
	List	   *context = NIL;

	initStringInfo(&sql);
	initStringInfo(&foreign_relname);

	/*
	 * First of all, determine which qual can be pushed down.
	 *
	 * The expressions which satisfy is_foreign_expr() are deparsed into WHERE
	 * clause of result SQL string, and they could be removed from PlanState
	 * to avoid duplicate evaluation at ExecScan().
	 *
	 * We never change the quals in the Plan node, because this execution might
	 * be for a PREPAREd statement, thus the quals in the Plan node might be
	 * reused to construct another PlanState for subsequent EXECUTE statement.
	 *
	 * We do this before deparsing SELECT clause because attributes which are
	 * not used in neither reltargetlist nor baserel->baserestrictinfo, quals
	 * evaluated on local, can be replaced with literal "NULL" in the SELECT
	 * clause to reduce overhead of tuple handling tuple and data transfer.
	 *
	 * Note: If we need to treat only simple foreiggn scan, we can remove quals
	 * which are going to evaluated on the foreign server from baserel in
	 * order to avoid overhead of duplicated evaluation.  But we treat joined
	 * foreign scan too, so we must keep baserestrictinfo as is for future
	 * evaluation.  Such optimization code fragmentations are in NOT_USED area.
	 */
	if (baserel->baserestrictinfo != NIL)
	{
		ListCell   *lc;
		List	   *attrs;
#ifdef NOT_USED
		List	   *local_qual = NIL;
#endif

		foreach (lc, baserel->baserestrictinfo)
		{
			RestrictInfo   *ri = (RestrictInfo *) lfirst(lc);

			/*
			 * We need to know which attributes are used in qual evaluated
			 * on the local server, because they should be listed in the
			 * SELECT clause of remote query.  We can ignore attributes
			 * which are referenced only in ORDER BY/GROUP BY clause because
			 * such attributes has already been kept in reltargetlist.
			 */
			attrs = pull_var_clause((Node *) ri->clause,
									PVC_RECURSE_AGGREGATES,
									PVC_RECURSE_PLACEHOLDERS);
			attr_used = list_union(attr_used, attrs);

			/* Determine whether the qual can be pushed down or not. */
			if (is_foreign_expr(root, baserel, ri->clause))
				foreign_expr = lappend(foreign_expr, ri->clause);
#ifdef NOT_USED
			else
			{
				/*
				 * Save the RestrictInfo to replace baserel->baserestrictinfo
				 * afterward.
				 */
				local_qual = lappend(local_qual, ri);
			}
#endif
		}

#ifdef NOT_USED
		/*
		 * Remove quals which are going to evaluated on the foreign server to
		 * avoid overhead of duplicated evaluation.
		 */
		baserel->baserestrictinfo = local_qual;
#endif
	}

	/*
	 * Determine foreign relation's qualified name.  This is necessary for
	 * FROM clause and SELECT clause.
	 */
	getQualifiedRelationName(relid, &foreign_relname);

	/*
	 * Create context for the deparse.
	 * We need multi-relation context without PlanState, and alias of each
	 * RangeTblEntry should be modified to be valid on remote side.
	 * 
	 * We skip first element of simple_rel_array.
	 */
	for (i = 1; i < root->simple_rel_array_size; i++)
	{
		RangeTblEntry *rte = copyObject(root->simple_rte_array[i]);
		rtable = lappend(rtable, rte);

		/*
		 * The alias name which is used in the local query is also used in
		 * remote query to reference proper column.
		 */
		if (i == baserel->relid)
			appendStringInfo(&foreign_relname, " %s", rte->eref->aliasname);
	}
	context = deparse_context_for_rangetblentries(rtable);

	/*
	 * deparse SELECT clause
	 *
	 * List attributes which are in either target list or local restriction.
	 * Unused attributes are replace with literal "NULL" for optimization.
	 */
	appendStringInfo(&sql, "SELECT ");
	attr_used = list_union(attr_used, baserel->reltargetlist);
	first = true;
	for (attr = 1; attr <= baserel->max_attr; attr++)
	{
		Var		   *var = NULL;
		ListCell   *lc;

		if (!first)
			appendStringInfo(&sql, ", ");
		first = false;

		/*
		 * We use linear search here, but it wouldn't be problem since
		 * attr_used seems to not become so large.
		 */
		foreach (lc, attr_used)
		{
			var = lfirst(lc);
			if (var->varattno == attr)
				break;
			var = NULL;
		}
		if (var != NULL)
			appendStringInfo(&sql, "%s",
				deparse_expression_pg((Node *) var, context, true, false));
		else
			appendStringInfo(&sql, "NULL");
	}

	/*
	 * deparse FROM clause
	 */
	appendStringInfo(&sql, " FROM %s", foreign_relname.data);

	/*
	 * deparse WHERE clause
	 */
	if (foreign_expr != NIL)
	{
		Node	   *node;

		node = (Node *) make_ands_explicit(foreign_expr);
		appendStringInfo(&sql, " WHERE %s ",
			deparse_expression_pg(node, context, true, false));
		list_free(foreign_expr);
		foreign_expr = NIL;
	}

	return sql.data;
}

/*
 * Returns true if expr is safe to be evaluated on the foreign server.
 */
static bool
is_foreign_expr(PlannerInfo *root, RelOptInfo *baserel, Expr *expr)
{
	foreign_executable_cxt	context;
	context.root = root;
	context.foreignrel = baserel;

	/*
	 * An expression which includes any mutable function can't be pushed down
	 * because it's result is not stable.  For example, pushing now() down to
	 * remote side would cause confusion from the clock offset.
	 * If we have routine mapping infrastructure in future release, we will be
	 * able to choose function to be pushed down in finer granularity.
	 */
	if (contain_mutable_functions((Node *) expr))
		return false;

	/*
	 * Check that the expression consists of nodes which are known as safe to
	 * be pushed down.
	 */
	if (foreign_expr_walker((Node *) expr, &context))
		return false;

	return true;
}

/*
 * Return true if node includes any node which is not known as safe to be
 * pushed down.
 */
static bool
foreign_expr_walker(Node *node, foreign_executable_cxt *context)
{
	if (node == NULL)
		return false;

	switch (nodeTag(node))
	{
		case T_Const:
		case T_ArrayExpr:
		case T_BoolExpr:
		case T_NullTest:
		case T_DistinctExpr:
		case T_ScalarArrayOpExpr:
			/*
			 * These type of nodes are known as safe to be pushed down.
			 * Of course the subtre of the node, if any, should be checked
			 * continuously at the tail of this function.
			 */
			break;
		case T_Param:
			/*
			 * Only external parameters can be pushed down.:
			 */
			{
				if (((Param *) node)->paramkind != PARAM_EXTERN)
					return true;
			}
			break;
		case T_OpExpr:
			/*
			 * Operators which use non-immutable function can't be pushed down.
			 */
			{
				OpExpr	   *oe = (OpExpr *) node;

				if (!is_proc_remotely_executable(oe->opfuncid))
					return true;

				/* operands are checked later */
			}
			break;
		case T_FuncExpr:
			/*
			 * Non-immutable functions can't be pushed down.
			 */
			{
				FuncExpr   *fe = (FuncExpr *) node;

				if (!is_proc_remotely_executable(fe->funcid))
					return true;

				/* operands are checked later */
			}
			break;
		case T_Var:
			/*
			 * Var can be pushed down if it is in the foreign table.
			 * XXX Var of other relation can be here?
			 */
			{
				Var	   *var = (Var *) node;
				foreign_executable_cxt *f_context;

				f_context = (foreign_executable_cxt *) context;
				if (var->varno != f_context->foreignrel->relid ||
					var->varlevelsup != 0)
					return true;
			}
			break;
		default:
			{
				ereport(DEBUG3,
						(errmsg("expression is too complex"),
						 errdetail("%s", nodeToString(node))));
				return true;
			}
			break;
	}

	return expression_tree_walker(node, foreign_expr_walker, context);
}

/*
 * Return true if func is known as safe to be pushed down if all of the
 * arguments are also known as safe.
 */
static bool
is_proc_remotely_executable(Oid procid)
{
	/*
	 * User-defined procedures can't be pushed down.
	 */
	if (procid >= FirstNormalObjectId)
		return false;

	return true;
}

/*
 * Deparse join representation into SQL statement which suits for remote
 * PostgreSQL server.  See also deparseSql().
 */
char *
deparseJoinSql(Oid serverid,
			   PlannerInfo *root,
			   RelOptInfo *joinrel,
			   JoinType jointype,
			   SpecialJoinInfo *sjinfo,
			   Path *outer_path,
			   Path *inner_path,
			   List *restrict_clauses,
			   List *pathkeys)
{
	int				i;
	List		   *rtables = NIL;
	List		   *context;
	StringInfoData	sql;
	StringInfoData	where;
	bool			first;
	ListCell	   *lc;
	Bitmapset	   *bms;

	/*
	 * First of all, check that this join can be pushed down.
	 *
	 * There are some redundancy such as duplicated attribute loop, but this
	 * is necessary to avoid memory leak.
	 */
	foreach(lc, joinrel->reltargetlist)
	{
		Var	   *var = (Var *) lfirst(lc);
		if (var->varattno < 0)
			return NULL;
	}

	/* Create context used for deparsing this join. */
	for (i = 1; i < root->simple_rel_array_size; i++)
	{
		rtables = lappend(rtables, copyObject(root->simple_rte_array[i]));
	}
	context = deparse_context_for_rangetblentries(rtables);

	/* Initialize buffer. */
	initStringInfo(&sql);

	/*
	 * deparse SELECT clause
	 */
	appendStringInfo(&sql, "SELECT ");
	first = true;
	foreach(lc, joinrel->reltargetlist)
	{
		Var	   *var = (Var *) lfirst(lc);
		appendStringInfo(&sql, "%s%s", first ? "" : ", ",
			deparse_expression_pg((Node *) var, context, true, false));
		first = false;
	}

	/*
	 * deparse FROM clause
	 */
	appendStringInfo(&sql, " FROM ");
	bms = bms_copy(joinrel->relids);
	first = true;
	while (!bms_is_empty(bms))
	{
		int relid = bms_first_member(bms);
		RangeTblEntry	   *rte = root->simple_rte_array[relid];

		if (!first)
			appendStringInfo(&sql, ", ");
		getQualifiedRelationName(rte->relid, &sql);
		appendStringInfo(&sql, " %s", rte->eref->aliasname);
		first = false;
	}

	/*
	 * deparse WHERE clause
	 *
	 * Note that we deparse separately because this join might have no
	 * qualifier; this might be a cartesian product.
	 */
	initStringInfo(&where);
	first = true;
	foreach(lc, restrict_clauses)
	{
		RestrictInfo   *ri  = (RestrictInfo *) lfirst(lc);

		if (!first)
			appendStringInfo(&where, " AND ");
		appendStringInfo(&where, "%s",
			deparse_expression_pg((Node *) ri->clause, context, true, false));
		first = false;
	}
	if (outer_path != NULL)
		append_where_clause(&where, context, outer_path, &first);
	if (inner_path != NULL)
		append_where_clause(&where, context, inner_path, &first);
	/* Append WHERE clause if any. */
	if (where.data[0] != '\0')
		appendStringInfo(&sql, " WHERE %s", where.data);

	/* return constructed SQL statement. */
	return sql.data;
}

/*
 * Append WHERE clause elements to buf.
 */
static void
append_where_clause(StringInfo buf, List *context, Path *path, bool *first)
{
	List		   *restrictinfo;
	ListCell	   *lc;
	Path		   *outer_path = NULL;
	Path		   *inner_path = NULL;

	/* Determine that which restrictinfo is used. */
	if (IsA(path, ForeignPath))
	{
		restrictinfo = path->parent->baserestrictinfo;
	}
	else if (IsA(path, ForeignJoinPath))
	{
		JoinPath	   *joinpath = (JoinPath *) path;
		restrictinfo = joinpath->joinrestrictinfo;
		outer_path = joinpath->outerjoinpath;
		inner_path = joinpath->innerjoinpath;
	}

	/* Deparse restrict expressions. */
	foreach(lc, restrictinfo)
	{
		RestrictInfo   *ri  = (RestrictInfo *) lfirst(lc);

		if (!*first)
			appendStringInfo(buf, " AND ");
		appendStringInfo(buf, "%s",
			deparse_expression_pg((Node *) ri->clause, context, true, false));
		*first = false;
	}

	/* Use children's restrictinfo recursively, if any. */
	if (outer_path != NULL)
		append_where_clause(buf, context, outer_path, first);
	if (inner_path != NULL)
		append_where_clause(buf, context, inner_path, first);
}
