/*-------------------------------------------------------------------------
 *
 * pgsql_fdw.c
 *		  foreign-data wrapper for remote PostgreSQL servers.
 *
 * Copyright (c) 2011, PostgreSQL Global Development Group
 *
 * IDENTIFICATION
 *		  pgsql_fdw/pgsql_fdw.c
 *
 *-------------------------------------------------------------------------
 */
#include "postgres.h"
#include "fmgr.h"

#include "catalog/pg_foreign_server.h"
#include "catalog/pg_foreign_table.h"
#include "commands/explain.h"
#include "foreign/fdwapi.h"
#include "funcapi.h"
#include "miscadmin.h"
#include "utils/lsyscache.h"
#include "utils/rel.h"

#include "pgsql_fdw.h"
#include "connection.h"

PG_MODULE_MAGIC;

/*
 * Default fetch count for cursor mode.  This can be override by fetch_count
 * FDW option.
 */
#define DEFAULT_FETCH_COUNT			1000

/*
 * Default minimum estimated row count for cursor mode.  This can be override
 * by min_cursor_rows FDW option.
 * */
#define DEFAULT_MIN_CURSOR_ROWS		1000

/*
 * Cursors which are used together in a local query require different name, so
 * we use simple incremental name for that purpose.  We don't care wrap around
 * of cursor_id because it's hard to imagine that 2^32 cursors are used in a
 * query.
 */
#define	CURSOR_NAME_FORMAT		"pgsql_fdw_cursor_%u"
static uint32 cursor_id = 0;

/*
 * Index of FDW-private items stored in FdwPlan.
 */
enum FdwPrivateIndex {
	FdwPrivateSelectSql,

	/* Items for cursor mode */
	FdwPrivateDeclareSql,
	FdwPrivateFetchSql,
	FdwPrivateResetSql,
	FdwPrivateCloseSql,

	/* # of elements stored in the list fdw_private */
	FdwPrivateNum,
};

/*
 * This macro can be used in the executor to determine whether the scan is
 * using CURSOR or not.
 */
#define USE_CURSOR(fdwplan)	\
	(list_length((fdwplan)->fdw_private) > FdwPrivateDeclareSql)

/*
 * Describes an execution state of a foreign scan against a foreign table
 * using pgsql_fdw. 
 */
typedef struct PgsqlFdwExecutionState
{
	FdwPlan	   *fdwplan;		/* FDW-specific planning information */
	PGconn	   *conn;			/* connection for the scan */

	Oid		   *param_types;	/* type array of external parameter */
	const char **param_values;	/* value array of external parameter */

	int			attnum;			/* # of non-dropped attribute */
	char	  **col_values;		/* column value buffer */
	AttInMetadata *attinmeta;	/* attribute metadata */

	PGresult   *res;			/* result of the scan, partial in cursor mode */
	int			nextrow;		/* # of returned tuple from res */
	bool		cursor_opened;	/* true if cursor is opened */
} PgsqlFdwExecutionState;

/*
 * SQL functions
 */
extern Datum pgsql_fdw_handler(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(pgsql_fdw_handler);

/*
 * FDW callback routines
 */
static FdwPlan *pgsqlPlanForeignScan(Oid foreigntableid,
									 PlannerInfo *root,
									 RelOptInfo *baserel);
static void pgsqlExplainForeignScan(ForeignScanState *node, ExplainState *es);
static void pgsqlBeginForeignScan(ForeignScanState *node, int eflags);
static TupleTableSlot *pgsqlIterateForeignScan(ForeignScanState *node);
static void pgsqlReScanForeignScan(ForeignScanState *node);
static void pgsqlEndForeignScan(ForeignScanState *node);
static FdwPlan * pgsqlPlanForeignJoin(Oid serverid,
									  PlannerInfo *root,
									  RelOptInfo *joinrel,
									  JoinType jointype,
									  SpecialJoinInfo *sjinfo,
									  Path *outer_path,
									  Path *inner_path,
									  List *restrict_clauses,
									  List *pathkeys);

/*
 * Helper functions
 */
static void estimate_costs(PlannerInfo *root,
						   RelOptInfo *baserel,
						   const char *sql,
						   Cost *startup_cost,
						   Cost *total_cost);
static void execute_query(ForeignScanState *node);
static void fetch_result(ForeignScanState *node);
static void store_result(TupleTableSlot *slot, PgsqlFdwExecutionState *festate);

/*
 * Foreign-data wrapper handler function: return a struct with pointers
 * to my callback routines.
 */
Datum
pgsql_fdw_handler(PG_FUNCTION_ARGS)
{
	FdwRoutine *fdwroutine = makeNode(FdwRoutine);

	fdwroutine->PlanForeignScan = pgsqlPlanForeignScan;
	fdwroutine->ExplainForeignScan = pgsqlExplainForeignScan;
	fdwroutine->BeginForeignScan = pgsqlBeginForeignScan;
	fdwroutine->IterateForeignScan = pgsqlIterateForeignScan;
	fdwroutine->ReScanForeignScan = pgsqlReScanForeignScan;
	fdwroutine->EndForeignScan = pgsqlEndForeignScan;
	fdwroutine->PlanForeignJoin = pgsqlPlanForeignJoin;

	PG_RETURN_POINTER(fdwroutine);
}

/*
 * pgsqlPlanForeignScan
 *		Create a FdwPlan for a scan on the foreign table
 */
static FdwPlan *
pgsqlPlanForeignScan(Oid foreigntableid,
					PlannerInfo *root,
					RelOptInfo *baserel)
{
	char	   *sql;
	FdwPlan	   *fdwplan;
	List	   *fdw_private = NIL;
	const char *min_cursor_rows_str;
	int	min_cursor_rows = DEFAULT_MIN_CURSOR_ROWS;

	/* Construct FdwPlan with cost estimates */
	fdwplan = makeNode(FdwPlan);

	sql = deparseSql(foreigntableid, root, baserel);
	elog(DEBUG3, "SCAN SQL: [%s]", sql);
	estimate_costs(root, baserel, sql,
				   &fdwplan->startup_cost, &fdwplan->total_cost);

	/*
	 * Store plain SELECT statement in private area of FdwPlan.  This will be
	 * used for executing remote query and explaining scan.
	 */
	fdw_private = list_make1(makeString(sql));

	/*
	 * We use SQL-level cursor when the result seems to be large, to limit
	 * memory usage for the result set. Exceptionally we use simple SELECT if
	 * min_cursor_rows is set to 0.
	 */
	min_cursor_rows_str = GetFdwOption(foreigntableid, "min_cursor_rows");
	if (min_cursor_rows_str != NULL)
		min_cursor_rows = strtol(min_cursor_rows_str, NULL, 10);
	elog(DEBUG1,
		 "relid=%u min_cursor_rows_str=%d mode=%s",
		 foreigntableid,
		 min_cursor_rows,
		 baserel->rows >= min_cursor_rows ? "cursor" : "select");

	if (min_cursor_rows != 0 && baserel->rows >= min_cursor_rows)
	{
		char			name[128];	/* must be larger than format + 10 */
		StringInfoData	cursor;
		const char	   *fetch_count_str;
		int				fetch_count = DEFAULT_FETCH_COUNT;

		/* Use specified fetch_count instead of default value, if any. */
		fetch_count_str = GetFdwOption(foreigntableid, "fetch_count");
		if (fetch_count_str != NULL)
			fetch_count = strtol(fetch_count_str, NULL, 10);
		elog(DEBUG1,
			 "relid=%u fetch_count=%d",
			 foreigntableid,
			 fetch_count);

		/* We store some more information in FdwPlan to pass them beyond the
		 * boundary between planner and executor.  Finally FdwPlan using cursor
		 * would hold items below:
		 *
		 * 1) plain SELECT statement (already added above)
		 * 2) SQL statement used to declare cursor
		 * 3) SQL statement used to fetch rows from cursor
		 * 4) SQL statement used to reset cursor 
		 * 5) SQL statement used to close cursor
		 *
		 * These items are indexed with the enum FdwPrivateIndex, so an item
		 * can be accessed directly via list_nth().  For example of FETCH
		 * statement:
		 *      list_nth(fdw_private, FdwPrivateFetchSql)
		 */

		/* Construct cursor name from sequential value */
		sprintf(name, CURSOR_NAME_FORMAT, cursor_id++);

		/* Construct statement to declare cursor */
		initStringInfo(&cursor);
		appendStringInfo(&cursor, "DECLARE %s SCROLL CURSOR FOR %s", name, sql);
		fdw_private = lappend(fdw_private, makeString(cursor.data));

		/* Construct statement to fetch rows from cursor */
		initStringInfo(&cursor);
		appendStringInfo(&cursor, "FETCH %d FROM %s", fetch_count, name);
		fdw_private = lappend(fdw_private, makeString(cursor.data));

		/* Construct statement to reset cursor */
		initStringInfo(&cursor);
		appendStringInfo(&cursor, "MOVE ABSOLUTE 0 FROM %s", name);
		fdw_private = lappend(fdw_private, makeString(cursor.data));

		/* Construct statement to close cursor */
		initStringInfo(&cursor);
		appendStringInfo(&cursor, "CLOSE %s", name);
		fdw_private = lappend(fdw_private, makeString(cursor.data));
	}

	/* Store FDW private information into FdwPlan */
	fdwplan->fdw_private = fdw_private;

	return fdwplan;
}

/*
 * pgsqlExplainForeignScan
 *		Produce extra output for EXPLAIN
 */
static void
pgsqlExplainForeignScan(ForeignScanState *node, ExplainState *es)
{
	FdwPlan	   *fdwplan;
	char	   *sql;

	fdwplan = ((ForeignScan *) node->ss.ps.plan)->fdwplan;
	if (USE_CURSOR(fdwplan))
		sql = strVal(list_nth(fdwplan->fdw_private, FdwPrivateDeclareSql));
	else
		sql = strVal(list_nth(fdwplan->fdw_private, FdwPrivateSelectSql));
	ExplainPropertyText("Remote SQL", sql, es);
}

/*
 * pgsqlBeginForeignScan
 *		Initiate access to the file by creating CopyState
 */
static void
pgsqlBeginForeignScan(ForeignScanState *node, int eflags)
{
	PgsqlFdwExecutionState *festate;
	PGconn		   *conn;
	ForeignServer  *server;
	UserMapping	   *user;
	TupleTableSlot *slot = node->ss.ss_ScanTupleSlot;

	/*
	 * Do nothing in EXPLAIN (no ANALYZE) case.  node->fdw_state stays NULL.
	 */
	if (eflags & EXEC_FLAG_EXPLAIN_ONLY)
		return;

	/*
	 * Save state in node->fdw_state.  We must save enough information to call
	 * BeginCopyFrom() again.
	 */
	festate = (PgsqlFdwExecutionState *) palloc(sizeof(PgsqlFdwExecutionState));
	festate->fdwplan = ((ForeignScan *) node->ss.ps.plan)->fdwplan;

	/*
	 * Get connection to the foreign server.  Connection manager would
	 * establish new connection if necessary.
	 */
	server = GetForeignServer(((ForeignScan *) node->ss.ps.plan)->serverid);
	user = GetUserMapping(GetOuterUserId(), server->serverid);
	conn = GetConnection(server, user);
	festate->conn = conn;

	/* Result will be filled in first Iterate call. */
	festate->res = NULL;
	festate->nextrow = 0;
	festate->cursor_opened = false;

	/* Allocate buffers for column values. */
	{
		TupleDesc	tupdesc = slot->tts_tupleDescriptor;
		festate->col_values = palloc(sizeof(char *) * tupdesc->natts);
		festate->attinmeta = TupleDescGetAttInMetadata(tupdesc);
	}

	/* Allocate buffers for query parameters. */
	{
		ParamListInfo	params = node->ss.ps.state->es_param_list_info;
		int				numParams = params ? params->numParams : 0;

		if (numParams > 0)
		{
			festate->param_types = palloc0(sizeof(Oid) * numParams);
			festate->param_values = palloc0(sizeof(char *) * numParams);
		}
		else
		{
			festate->param_types = NULL;
			festate->param_values = NULL;
		}
	}


	/* Store FDW-specific state into ForeignScanState */
	node->fdw_state = (void *) festate;

	return;
}

/*
 * pgsqlIterateForeignScan
 *   - retrieve next row from the result set, or clear tuple slot to indicate
 *     EOF.
 */
static TupleTableSlot *
pgsqlIterateForeignScan(ForeignScanState *node)
{
	PgsqlFdwExecutionState *festate;
	TupleTableSlot *slot = node->ss.ss_ScanTupleSlot;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;

	/*
	 * If this is the first call after Begin, we need to execute remote query.
	 * If the query needs cursor, we declare a cursor at first call and fetch
	 * from it in later calls.
	 */
	if (festate->res == NULL && !festate->cursor_opened)
		execute_query(node);

	/*
	 * If enough tuples are left in PGresult, just return next tuple from it.
	 */
	if (festate->nextrow < PQntuples(festate->res))
	{
		store_result(slot, festate);
		return slot;
	}

	/* If the scan doesn't use cursor, the scan has been done. */
	if (!USE_CURSOR(festate->fdwplan))
	{
		ExecClearTuple(slot);
		return slot;
	}

	/*
	 * Here we need to clear partial result and fetch next bunch of tuples from
	 * from the cursor for the scan.  If the fetch returns no tuple, the scan
	 * has reached the end.
	 */
	fetch_result(node);
	if (festate->nextrow < PQntuples(festate->res))
	{
		store_result(slot, festate);
		return slot;
	}
	else
	{
		ExecClearTuple(slot);
		return slot;
	}
}

/*
 * pgsqlReScanForeignScan
 *   - Restart this scan by resetting fetch location.
 */
static void
pgsqlReScanForeignScan(ForeignScanState *node)
{
	List	   *fdw_private;
	char	   *sql;
	PGconn	   *conn;
	PGresult   *res;
	PgsqlFdwExecutionState *festate;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;

	/* Nothing to do when result has not been fetched yet. */
	if (festate->res == NULL)
		return;

	PQclear(festate->res);
	festate->res = NULL;
	festate->nextrow = 0;

	/* Reset cursor */
	if (USE_CURSOR(festate->fdwplan))
	{
		fdw_private = festate->fdwplan->fdw_private;
		conn = festate->conn;
		sql = strVal(list_nth(fdw_private, FdwPrivateResetSql));
		res = PQexec(conn, sql);
		if (res == NULL || PQresultStatus(res) != PGRES_COMMAND_OK)
		{
			char *msg;

			msg = pstrdup(PQerrorMessage(conn));
			PQclear(res);
			ereport(ERROR,
					(errmsg("could not rewind cursor"),
					 errdetail("%s", msg),
					 errhint("%s", sql)));
		}
	}
}

/*
 * pgsqlEndForeignScan
 *		Finish scanning foreign table and dispose objects used for this scan
 */
static void
pgsqlEndForeignScan(ForeignScanState *node)
{
	List	   *fdw_private;
	char	   *sql;
	PGconn	   *conn;
	PGresult   *res;
	PgsqlFdwExecutionState *festate;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;

	/* if festate is NULL, we are in EXPLAIN; nothing to do */
	if (festate == NULL)
		return;

	/* Discard fetch results */
	PQclear(festate->res);

	/* Close cursor */
	if (USE_CURSOR(festate->fdwplan) && festate->cursor_opened)
	{
		fdw_private = festate->fdwplan->fdw_private;
		conn = festate->conn;
		sql = strVal(list_nth(fdw_private, FdwPrivateCloseSql));
		res = PQexec(conn, sql);
		if (res == NULL || PQresultStatus(res) != PGRES_COMMAND_OK)
		{
			char *msg;

			msg = pstrdup(PQerrorMessage(conn));
			PQclear(res);
			ereport(ERROR,
					(errmsg("could not close cursor"),
					 errdetail("%s", msg),
					 errhint("%s", sql)));
		}
	}

	ReleaseConnection(festate->conn);
}

/*
 * Estimate costs of scanning a foreign table.
 */
static void
estimate_costs(PlannerInfo *root, RelOptInfo *baserel,
			   const char *sql,
			   Cost *startup_cost, Cost *total_cost)
{
	double			connection_cost;
	double			transfer_cost;

	connection_cost = 10.0;
	transfer_cost = 0.01;		/* transfer cost per byte */

	/* TODO Selectivity of quals pushed down should be considered. */

	*startup_cost += connection_cost;
	*total_cost += connection_cost;
	*total_cost += transfer_cost * baserel->width * baserel->tuples;
}

/*
 * Execute remote query with current parameters.
 */
static void
execute_query(ForeignScanState *node)
{
	FdwPlan		   *fdwplan;
	PgsqlFdwExecutionState *festate;
	ParamListInfo	params = node->ss.ps.state->es_param_list_info;
	int				numParams = params ? params->numParams : 0;
	Oid			   *types = NULL;
	const char	  **values = NULL;
	char		   *sql;
	int				required_result;
	PGconn		   *conn;
	PGresult	   *res;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;
	types = festate->param_types;
	values = festate->param_values;

	/*
	 * Construct parameter array in text format.  We don't release memory for
	 * the arrays explicitly, because the memory usage would not be very large,
	 * and anyway they will be released in context cleanup.
	 */
	if (numParams > 0)
	{
		int i;

		for (i = 0; i < numParams; i++)
		{
			types[i] = params->params[i].ptype;
			if (params->params[i].isnull)
				values[i] = NULL;
			else
			{
				Oid			out_func_oid;	
				bool		isvarlena;
				FmgrInfo	func;

				getTypeOutputInfo(types[i], &out_func_oid, &isvarlena);
				fmgr_info(out_func_oid, &func);
				values[i] = OutputFunctionCall(&func, params->params[i].value);
			}
		}
	}

	/*
	 * Execute remote query with parameters.
	 */
	conn = festate->conn;
	fdwplan = ((ForeignScan *) node->ss.ps.plan)->fdwplan;
	if (USE_CURSOR(fdwplan))
	{
		sql = strVal(list_nth(fdwplan->fdw_private, FdwPrivateDeclareSql));
		required_result = PGRES_COMMAND_OK;
	}
	else
	{
		sql = strVal(list_nth(fdwplan->fdw_private, FdwPrivateSelectSql));
		required_result = PGRES_TUPLES_OK;
	}
	res = PQexecParams(conn, sql, numParams, types, values, NULL, NULL, 0);

	/*
	 * If the query has failed, reporting details is enough here.
	 * Connection(s) which are used by this query (at least used by pgsql_fdw)
	 * will be cleaned up by the foreign connection manager.
	 */
	if (res == NULL || PQresultStatus(res) != required_result)
	{
		char *msg;

		msg = pstrdup(PQerrorMessage(conn));
		PQclear(res);
		ereport(ERROR,
				(errmsg("could not execute foreign query"),
				 errdetail("%s", msg),
				 errhint("%s", sql)));
	}

	/* Fetch first bunch of result if we using cursor. */
	if (USE_CURSOR(fdwplan))
	{
		festate->cursor_opened = true;
		fetch_result(node);
	}
	else
	{
		festate->res = res;
		festate->nextrow = 0;
	}
}

/*
 * Fetch next partial result from remote server.
 */
static void
fetch_result(ForeignScanState *node)
{
	PgsqlFdwExecutionState *festate;
	List		   *fdw_private;
	char		   *sql;
	PGconn		   *conn;
	PGresult	   *res;

	festate = (PgsqlFdwExecutionState *) node->fdw_state;

	/* Clear current result */
	PQclear(festate->res);
	festate->res = NULL;
	festate->nextrow = 0;

	/* retrieve information for fetching result. */
	fdw_private = festate->fdwplan->fdw_private;
	sql = strVal(list_nth(fdw_private, FdwPrivateFetchSql));
	conn = festate->conn;
	res = PQexec(conn, sql);
	if (res == NULL || PQresultStatus(res) != PGRES_TUPLES_OK)
	{
		char *msg;

		msg = pstrdup(PQerrorMessage(conn));
		PQclear(res);
		ereport(ERROR,
				(errmsg("could not fetch rows from foreign server"),
				 errdetail("%s", msg),
				 errhint("%s", sql)));
	}

	/* save partial result and return the tuple on the head of it. */
	festate->res = res;
	festate->nextrow = 0;
}

/*
 * Create a HeapTuple from PGresult and store it into tuple slot.
 */
static void
store_result(TupleTableSlot *slot, PgsqlFdwExecutionState *festate)
{
	int					i;
	int					nfields;
	int					attnum;		/* number of non-dropped columns */
	Form_pg_attribute  *attrs;
	TupleDesc	tupdesc = slot->tts_tupleDescriptor;

	nfields = PQnfields(festate->res);
	attrs = tupdesc->attrs;

	/* count non-dropped columns */
	for (attnum = 0, i = 0; i < tupdesc->natts; i++)
		if (!attrs[i]->attisdropped)
			attnum++;

	/* check result and tuple descriptor have the same number of columns */
	if (attnum > 0 && attnum != nfields)
		ereport(ERROR,
				(errcode(ERRCODE_DATATYPE_MISMATCH),
				 errmsg("remote query result rowtype does not match "
						"the specified FROM clause rowtype"),
				 errdetail("expected %d, actual %d", attnum, nfields)));

	/* put a tuples into the slot */
	{
		int			j;
		HeapTuple	tuple;

		for (i = 0, j = 0; i < tupdesc->natts; i++)
		{
			/* skip dropped columns. */
			if (attrs[i]->attisdropped)
			{
				festate->col_values[i] = NULL;
				continue;
			}

			if (PQgetisnull(festate->res, festate->nextrow, j))
				festate->col_values[i] = NULL;
			else
				festate->col_values[i] = PQgetvalue(festate->res, festate->nextrow, j);
			j++;
		}

		/*
		 * Build the tuple and put it into the slot.
		 * We don't have to free the tuple explicitly because it's been
		 * allocated in the per-tuple context.
		 */
		tuple = BuildTupleFromCStrings(festate->attinmeta, festate->col_values);
		ExecStoreTuple(tuple, slot, InvalidBuffer, false);
	}

	/* Advance row pointer. */
	festate->nextrow++;
}

/*
 * Create FdwPlan for a foreignjoin node.
 */
static FdwPlan *
pgsqlPlanForeignJoin(Oid serverid,
					 PlannerInfo *root,
					 RelOptInfo *joinrel,
					 JoinType jointype,
					 SpecialJoinInfo *sjinfo,
					 Path *outer_path,
					 Path *inner_path,
					 List *restrict_clauses,
					 List *pathkeys)
{
	char	   *sql;
	FdwPlan	   *fdwplan;

	Assert(joinrel->serverid != InvalidOid);
	Assert(IsA(outer_path, ForeignJoinPath) || IsA(outer_path, ForeignPath));
	Assert(IsA(inner_path, ForeignJoinPath) || IsA(outer_path, ForeignPath));

	/*
	 * In first version, consider only inner join, because OUTER JOIN may
	 * populates result rows more than retrieving each relation separately.
	 */
	if (jointype != JOIN_INNER)
		return NULL;

	/*
	 * Deparse join represention into SQL statement which retrieves result of
	 * the join, and create FdwPlan with that SQL statement.
	 */
	sql = deparseJoinSql(serverid, root, joinrel, jointype, sjinfo,
						 outer_path, inner_path, restrict_clauses, pathkeys);
	elog(DEBUG3, "JOIN SQL: [%s]", sql);
	if (sql == NULL)
		return NULL;

	fdwplan = makeNode(FdwPlan); 
	fdwplan->startup_cost = 0.0;
	fdwplan->total_cost = 0.0;
	fdwplan->fdw_private = list_make1(makeString(sql));

	return fdwplan;
}

